
from .biocham import import_biocham
from .cadbiom import import_cadbiom
from .sbgnpd import import_sbgnpd
from .minibn import import_minibn

CONVERTERS = {
    "biocham": import_biocham,
    "cadbiom": import_cadbiom,
    "sbgnpd": import_sbgnpd,
}





import subprocess
from colomoto.setup_helper import setup, PKG

__pint_required__ = "2018-07-10"
__PINT_VERSION__ = None

def check_pint_version():
    global __PINT_VERSION__
    version = subprocess.check_output(["pint-config", "version"]).decode()
    __PINT_VERSION__ = version.strip()
    return __pint_required__ <= __PINT_VERSION__

if __name__ == "__main__":
    setup({"pkg": "colomoto/pint", "check_progs": ["pint-config", "pint-reach"],
                "check_install": check_pint_version},
        {"pkg": "colomoto/mole", "check_progs": ["mole"]},
        PKG["clingo"],
        PKG["ginsim"],
        PKG["its"],
        PKG["nusmv"])



"""
When loaded, the `pypint` module will perform the follow tasks:

* add Pint third-party binary path (`bin` subdirectory of
  ``pint-config share-path``) to the ``PATH`` environment variable.
* if in IPython, displays the version of Pint binaries, and executes
  :py:func:`.ipython_install`.
"""
__version__ = "1.6.1"

import os
import shutil
import subprocess
import sys

if sys.version_info[0] < 3:
    raise Exception("python >= 3 is required")

from .cfg import *
from .types import *
from .model import *
from .ui import *
from .tools import *
from .utils import *

import colomoto.setup_helper
import pypint_setup

__SETUP_DONE = False

__PINT_VERSION__ = None

def check_pint():
    if shutil.which("pint-config") is None:
        raise Exception("Pint binaries should be installed separately. " + \
            "See https://loicpauleve.name/pint/doc/#Binaries" + \
            " or run\n\tpython -m pypint_setup")

    if not pypint_setup.check_pint_version():
        raise Exception("Pint >= {0} is required. Please upgrade Pint binaries. " \
                .format(pypint_setup.__pint_required__) +  \
            "See https://loicpauleve.name/pint/doc/#Binaries" + \
            " or run\n\tpython -m pypint_setup")
    global __PINT_VERSION__
    __PINT_VERSION__ = pypint_setup.__PINT_VERSION__


def setup_environ():
    check_pint()
    global __SETUP_DONE
    share_path = subprocess.check_output(["pint-config", "share-path"]).decode()
    bin_path = os.path.join(share_path.strip(), "bin")
    os.environ["PATH"] = "%s:%s" % (bin_path, os.environ["PATH"])
    __SETUP_DONE = True

if not __SETUP_DONE:
    setup_environ()

if IN_IPYTHON:
    from colomoto_jupyter import jupyter_setup

    menu = [
        {"name":"Load model",
            "snippet":["an = pypint.load(\"filename_or_URL\")"]},
        {"name":"Upload model",
            "snippet":["an = pypint.load()"]},
        "---",
        {"name":"Model description",
            "sub-menu": [
            {"name": "Dependency graph",
                "snippet":['an.dependency_graph()']},
            {"name": "Model summary",
                "snippet":['an.summary()']}
            ]},
        {"name":"Export to file",
            "sub-menu": [
            {"name": "Pint native format (.an)",
                "snippet":['an.export("an")']},
            {"name": "NuSMV file (.smv)",
                "snippet":['an.export("nusmv")']},
            {"name": "Safe Petri net in PEP format (.ll)",
                "snippet":['an.export("pep")']},
            {"name": "Safe Petri net in ROMEO format (.xml)",
                "snippet":['an.export("romeo")']},
            ]},
        {"name":"Convert to tool",
            "sub-menu": [
            {"name": "ITS-Tools",
                "snippet":['itsm = an.to_its()']},
            {"name": "NuSMV",
                "snippet":['smv = an.to_nusmv()']},
            ]},
        "---",
        {"name":"Model transformation",
            "sub-menu": [
            {"name": "Change initial state",
                "snippet":['an.having(a=1,b=1)']},
            {"name": "Lock automata (mutations)",
                "snippet":['an.lock({"a":1,"b":0})']},
            {"name": "Disable local states",
                "snippet":['an.disable({"a":1,"b":1})']},
            {"name": "Goal-oriented reduction",
                "snippet":['an.reduce_for_goal("a=1")']}
            ]},
        "---",
        {"name":"Compute mutations for cutting goal reachability",
            "snippet":['an.oneshot_mutations_for_cut("a=1")']},
        {"name":"Compute cut sets of paths to goal",
            "snippet":['an.cutsets("a=1")']},
        {"name":"Compute bifurcation transitions from goal",
            "snippet":['an.bifurcations("a=1")']},
        {"name":"Verify reachability of goal",
            "snippet":['an.reachability("a=1")']},
        "---",
        {"name":"Local Causality Graph",
            "sub-menu": [
            {"name": "Full LCG", "snippet":['an.full_lcg()']},
            {"name": "Simple LCG for goal reachability over-approximation",
                "snippet":['an.simple_lcg("a=1")']},
            {"name": "Saturated LCG for goal reachability under-approximation",
                "snippet":['an.saturated_lcg("a=1")']},
            {"name": "Worth LCG for goal-oriented model reduction",
                "snippet":['an.worth_lcg("a=1")']}
            ]},
        "---",
        {"name":"State graph analysis",
            "sub-menu": [
            {"name": "Count reachable states",
                "snippet":['an.count_reachable_states()']},
            {"name": "Reachable state graph",
                "snippet":['an.reachable_stategraph()']},
            {"name": "Reachable attractors",
                "snippet":['an.reachable_attractors()']},
            {"name": "Fixpoints",
                "snippet":['an.fixpoints()']}
            ]},
        "---",
        {"name":"Goal specification",
            "sub-menu": [
            {"name": "Simple goal", "snippet":['"a=1"']},
            {"name": "Sub-state goal", "snippet":['"a=1,b=1"']},
            {"name": "Sequence of simple goals", "snippet":['"a=1","b=1"']},
            {"name": "Sequence of sub-state goals", "snippet":['"a=1,c=1","b=1,d=0"']},
            {"name": "Alternative goals", "snippet":['pypint.Goal("a=1")|pypint.Goal("b=1")']}
            ]},
        "---",
        {"name": "Documentation",
            "external-link": "https://loicpauleve.name/pint/doc/api.html"}
    ]
    toolbar = [
        {"name": "upload", "setup": {
            "icon": "fa-upload",
            "help": "Upload model",
            "handler": "action_upload_model"}},
        {"name": "enable-debug", "setup": {
            "help": "Enable debug",
            "handler": "action_enable_debug"}},
        {"name": "disable-debug", "setup": {
            "help": "Disable debug",
            "handler": "action_disable_debug"}},
    ]


    js_api = {
    "action_upload_model": """function() {
        var cell = Jupyter.notebook.get_selected_cell();
        cell.set_text('an = '+pypint_jsapi.module_alias+'.load()');
        cell.focus_editor();
    }""",
    "action_enable_debug": """function() {
        IPython.notebook.kernel.execute(pypint_jsapi.module_alias+".enable_dbg()");
        pypint_jsapi.debug_enabled(true);
    }""",
    "action_disable_debug": """function() {
        IPython.notebook.kernel.execute(pypint_jsapi.module_alias+".disable_dbg()");
        pypint_jsapi.debug_enabled(false);
    }""",
    "btn_enable_debug": "null",
    "btn_disable_debug": "null",
    "debug_enabled": """function(enabled) {
        if (enabled) {
            this.btn_enable_debug.hide();
            this.btn_disable_debug.show();
        } else {
            this.btn_enable_debug.show();
            this.btn_disable_debug.hide();
        }
    }""",
    "post_install_callback": """function() {
        this.btn_enable_debug = $("#pypint-toolbar > button[data-jupyter-action='pypint:enable-debug']");
        this.btn_disable_debug = $("#pypint-toolbar > button[data-jupyter-action='pypint:disable-debug']");
        this.btn_enable_debug[0].innerHTML = "enable debug";
        this.btn_disable_debug[0].innerHTML = "disable debug";
        this.debug_enabled(%s);
    }""" % (1 if CFG["dbg"] else 0)
    }

    jupyter_setup(__name__, label="Pint",
        color="red",
        menu=menu,
        toolbar=toolbar,
        js_api=js_api)





import os
import tempfile

from colomoto_jupyter import IN_IPYTHON
from colomoto_jupyter.sessionfiles import \
    output_dir, \
    new_output_file, \
    remove_output_files

CFG = {
    "dbg": False,
    "console": False,
}
"""
Python module configuation:

* `dbg`: enable debug output (see also :py:func:`.enable_dbg`,
  :py:func:`.disable_dbg`, :py:func:`.dbg`).
* `console`: assume console output
"""

__DATAFILES_DIR__ = os.path.join(os.path.dirname(__file__), "files")

def data_file(name):
    """
    Returns path to file `name` distributed with ``pypint``.
    """
    return os.path.join(__DATAFILES_DIR__, name)

__all__ = [
    "CFG",
    "data_file",
    "output_dir",
    "new_output_file",
    "remove_output_files",
    "IN_IPYTHON",
]




from io import StringIO
import json
import os
import subprocess
import tempfile

from functools import reduce

import networkx as nx

from colomoto.temporal_logics import *

from .cfg import *
from .types import *
from .ui import *
from .utils import *

if IN_IPYTHON:
    from IPython.display import FileLink

VALID_EXE = [
    "pint-export",
    "pint-lcg",
    "pint-mole",
    "pint-reach",
    "pint-sg",
    "pint-stable",
]

class PintProcessError(subprocess.CalledProcessError):
    """
    Exception raised when a Pint command fails.
    """
    def __str__(self):
        stderr = "\n%s" % self.stderr.decode() if self.stderr else self.output.decode()
        return "Command '%s' returned non-zero exit status %d%s" \
            % (" ".join(self.cmd), self.returncode, stderr)

def _run_tool(cmd, *args, input_model=None, reduce_for_goal=None, **run_opts):
    assert cmd in VALID_EXE
    args = list(args)
    args.insert(0, "--json-stdout")
    if "stdout" not in run_opts:
        run_opts["stdout"] = subprocess.PIPE
    if "stderr" not in run_opts:
        run_opts["stderr"] = subprocess.PIPE
    if "check" not in run_opts:
        run_opts["check"] = True

    assert (not reduce_for_goal or input_model)

    args = [a.to_pint() if isinstance(a, Goal) else a for a in args]

    args.insert(0, cmd)

    if reduce_for_goal:
        reduce_for_goal = Goal.from_arg(reduce_for_goal)
        pre_args = ["pint-export", "--reduce-for-goal",
                        reduce_for_goal.to_pint(), "--squeeze"]
        for a in reduce_for_goal.automata:
            pre_args += ["--squeeze-preserve", a]
        pre_kwargs = {}
        input_model.populate_popen_args(pre_args, pre_kwargs)
        pre_cmd = subprocess.Popen(pre_args, stdout=subprocess.PIPE)
        if "input" in pre_kwargs:
            pre_cmd.stdin.write(pre_kwargs["input"])
            pre_cmd.stdin.close()
        run_opts["stdin"] = pre_cmd.stdout
        dbg("Running command %s | %s" % (" ".join(pre_args), " ".join(args)))

    else:
        if input_model is not None:
            input_model.populate_popen_args(args, run_opts)

        dbg("Running command %s" % (" ".join(args)))
    try:
        return subprocess.run(args, **run_opts)
    except subprocess.CalledProcessError as e:
        # backward compatible 'raise e from None'
        e = PintProcessError(e.returncode, e.cmd, e.output, e.stderr)
        e.__cause__ = None
        raise e


format_alias = {
    "an": "dump",
}

format2ext = {
    "asp": "asp",
    "dump": "an",
    "nusmv": "smv",
    "pep": "ll",
    "pnml": "pnml",
    "romeo": "xml",
}
ext2format = dict([(j,i) for (i,j) in format2ext.items()])
ext2format["lp"] = "asp"

EXPORT_SUPPORTED_FORMATS = list(sorted(format2ext.keys()))
"""
Formats supported by :py:meth:`.Model.export` method.
"""

EXPORT_SUPPORTED_EXTENSIONS = list(sorted(ext2format.keys()))
"""
File extensions supported by :py:meth:`.Model.save_as` method:

* ``.an``: Pint native format
* ``.ll``: PEP format (Petri net)
* ``.pnml``: PNML format (Petri net)
* ``.smv``: NuSMV format
* ``.xml``: Romeo format (Petri net)
"""

__MODEL_TOOLS = []
def modeltool(f):
    __MODEL_TOOLS.append((f.__name__, f))
    return f

def EquipTools(cls):
    for name, func in __MODEL_TOOLS:
        setattr(cls, name, func)
    return cls

#
# pint-export
#

@modeltool
def export(self, format, output=None, reduce_for_goal=None, raw_args=None):
    """
    Export the AN model to the given `format`. If not provided the `output`
    filename is generated using :py:func:`.new_output_file` and the extension
    corresponding to the export format.

    e. seealso:: :py:data:`.EXPORT_SUPPORTED_FORMATS`, and similar method :py:meth:`.save_as`
    """
    format = format.lower()
    format = format_alias.get(format, format)
    assert format in format2ext
    args = ["-l", format]
    if raw_args is None:
        raw_args = []
    assert "-o" not in raw_args
    if not output:
        output = new_output_file(ext=format2ext[format])
    args += ["-o", output]
    _run_tool("pint-export", *args, *raw_args, reduce_for_goal=reduce_for_goal,
                input_model=self, stdout=None)
    if IN_IPYTHON:
        return FileLink(output)
    return output

@modeltool
def save_as(self, filename):
    """
    Save the AN model with its initial state to local file `filename`.
    The format is guessed from the extension of `filename`.

    .. seealso:: :py:data:`.EXPORT_SUPPORTED_EXTENSIONS`, and similar method :py:meth:`.export`

    Examples:

    >>> m.save_as("mymodel.an") # save to a local file in Pint native format
    >>> m.save_as("mymodel.ll") # export in PEP format (Petri net)
    """
    ext = file_ext(filename)
    assert ext in ext2format, "File extension is not supported"
    return export(self, ext2format[ext], output=filename)

@modeltool
def to_nusmv(self, skip_init=True, existential_ctx=True, reduce_for_goal=None):
    """
    TODO

    :keyword bool skip_init: do not generate an INIT statement.
    :keyword bool existential_ctx: if True (default),
        whenever the initial state is partial, the properties have to hold from
        at least one complete initial state;
        if False, the properties have to hold from all the possible complete initial
        states.
    :keyword reduce_for_goal: perform goal-oriented model reduction before
        exportation.
    :type reduce_for_goal: str or list(str) or .Goal
    """
    format = "nusmv"
    smvfile = new_output_file(ext=format2ext[format])
    mapfile = new_output_file(ext="json")

    raw_args = ["--mapfile", mapfile,
        "--existential-ctx" if existential_ctx else "--universal-ctx",
    ]
    if skip_init:
        raw_args += ["--no-init"]
    try:
        export(self, format, output=smvfile, reduce_for_goal=reduce_for_goal,
                raw_args=raw_args)
        with open(mapfile) as mf:
            bindings = json.load(mf)
    finally:
        os.unlink(mapfile)

    bindings = dict([(tuple(k),v) for k,v in bindings])
    from colomoto.modelchecking import ColomotoNuSMV
    return ColomotoNuSMV(smvfile, bindings.get)

@modeltool
def to_its(self, reduce_for_goal=None):
    """
    TODO

    :keyword reduce_for_goal: perform goal-oriented model reduction before
        exportation.
    :type reduce_for_goal: str or list(str) or .Goal
    """
    format = "romeo"
    outfile = new_output_file(ext=format2ext[format])
    mapfile = new_output_file(ext="json")

    raw_args = ["--mapfile", mapfile]
    try:
        export(self, format, output=outfile, reduce_for_goal=reduce_for_goal,
                raw_args=raw_args)
        with open(mapfile) as mf:
            bindings = json.load(mf)
    finally:
        os.unlink(mapfile)

    bindings = dict([(tuple(k),v) for k,v in bindings])
    from colomoto.modelchecking import ColomotoITS
    return ColomotoITS(outfile, bindings.get)


def _model_modification(self, args):
    output = new_output_file(ext="an")
    args += ["-o", output]
    _run_tool("pint-export", *args, input_model=self, stdout=None)
    from .model import FileModel
    return FileModel(output)


@modeltool
def simplify(self):
    """
    Simplify the local transitions by computing their prime implicants.

    :returns: a new :py:class:`.FileModel` instance for the reduced automata
        network.
    """
    return _model_modification(self, ["--simplify"])

@modeltool
def reduce_for_goal(self, goal=None, squeeze=True, squeeze_preserve=None, **kwgoal):
    """
    Compute the goal-oriented reduction of the automata network.
    The reduction removes local transitions that are guaranteed to never
    contribute to any minimal path from the initial state to the goal
    (a path is minimal if there exists no paths using a strict subset of
    transitions leading to the goal).

    The reduced models preserves existential reachability properties, as well as
    cut sets.

    The complexity of the method is polynomial in the number of local
    transitions and automata, and exponential in the size of the largest
    automata minus 1.

    :param goal: goal specification (e.g., ``"a=1"``)
    :type goal: str or list(str) or .Goal
    :keyword kwgoal: keywords for goal specification (instead of `goal` argument)
    :param bool squeeze: if ``True`` (default), unused automata and local states
        are removed. Warning: this can lead to local state renaming.
    :param str list squeeze_preserve:
        do not squeeze the given automata
    :returns: a new :py:class:`.FileModel` instance for the reduced automata
        network.

    Example:

    >>> redm = m.reduce_for_goal("g=1")
    """
    goal = Goal.from_arg(goal, **kwgoal)

    args = ["--reduce-for-goal", goal.to_pint()]
    if squeeze:
        args.append("--squeeze")
        if squeeze_preserve:
            for a in squeeze_preserve:
                args += ["--squeeze-preserve", a]
    fm = _model_modification(self, args)
    return fm

@modeltool
def inject_goal(self, goal=None, **kwgoal):
    """
    Whenever the goal is not a single local state, creates a new automaton which
    will reach is final state whenever the goal is complete.
    """
    goal = Goal.from_arg(goal, **kwgoal)
    if goal.is_simple_goal:
        return self, goal.local_state()

    fm = _model_modification(self, ["--inject-goal", goal.to_pint()])
    a = "_pint_goal"
    ls = (a, max(fm.local_states[a]))
    return fm, ls

@modeltool
def lock(self, substate={}, **kwstate):
    """
    Returns the AN where the specified automata are lock to the given local state.
    The initial state is modified accordingly.

    :param substate: local states to lock.
    :type substate: list(tuple(str,int)) or dict[str,int]
    :keyword kwstate: complementary specification of local states.
    :returns: a new :py:class:`.FileModel` instance for the modified AN.
    """
    lockstate = dict(**kwstate)
    if hasattr(substate, "items"):
        substate = substate.items()
    for a,i in substate:
        assert (isinstance(i, (str, int)) and \
            (not a in lockstate or lockstate[a] == i)), \
            "Cannot lock an automaton to different local states"
        lockstate[a] = i

    args = ["--lock", pint_of_localstates(lockstate.items())]

    return _model_modification(self, args)


@modeltool
def disable(self, local_states=[], **kwstate):
    """
    Returns the AN where all local transitions involving specified local states
    have been removed.

    :param local_states: local states to disable.
    :type local_states: list(tuple(str,int)) or dict[str,int]
    :keyword kwstate: complementary specification of local states.
    :returns: a new :py:class:`.FileModel` instance for the modified AN.
    """
    if hasattr(local_states, "items"):
        local_states = list(local_states.items())
    else:
        local_states = list(local_states)
    local_states += list(kwstate.items())

    args = ["--disable", pint_of_localstates(local_states)]

    return _model_modification(self, args)


#
# pint-reach
#

@modeltool
def cutsets(self, goal=None, maxsize=5, exclude=[], exclude_initial_state=False,
                exclude_goal_automata=True, timeout=None, **kwgoal):
    """
    Computes sets of local states which are used in all the paths from the
    initial state to `goal`:
    if all the local states in a cut-set are disabled (all related local
    transitions are removed), then it is impossible to reach `goal` from the
    initial state.

    Elements of the returned lists can be notably used as argument for methods
    :py:meth:`.Model.having` and :py:meth:`.disable`.

    :param goal: goal specification (e.g., ``"a=1"``)
    :type goal: str or list(str) or .Goal
    :keyword kwgoal: keywords for goal specification (instead of `goal` argument)
    :keyword int maxsize: maximal cardinality of a cut-set.
    :keyword set(str) exclude:
        set/list of automata to exclude from the solutions
    :keyword bool exclude_initial_state:
        if ``True``, cut-sets can not be composed of initial local states.
    :keyword bool exclude_goal_automata:
        exclude automata involved in the goal specification
    :param int timeout: command timeout in seconds
    :rtype: list(dict[str,int or int list])

    .. seealso:: method :py:meth:`.oneshot_mutations_for_cut`
    """
    goal = Goal.from_arg(goal, **kwgoal)

    args = []

    info("This computation is an *under-approximation*: returned cut-sets \
are all valid, but they may be non-minimal, and some cut-sets may be missed.")
    info("Limiting results to cut-sets with at most %s elements. Use `maxsize` argument to change." % maxsize)

    if exclude_initial_state:
        args.append("--no-init-cutsets")

    if isinstance(exclude, str):
        exclude = set([exclude])
    else:
        exclude = set(exclude)
    if exclude_goal_automata:
        exclude.update(goal.automata)
    if exclude:
        args += ["--ignore-automata", ",".join(['"%s"' % a for a in exclude])]

    cp = _run_tool("pint-reach", "--cutsets", str(maxsize), goal, *args,
                input_model=self, timeout=timeout)
    output = cp.stdout.decode()
    return json.loads(output)

@modeltool
def oneshot_mutations_for_cut(self, goal=None, maxsize=5,
        exclude=[],
        exclude_goal_automata=True,
        timeout=None,
        **kwgoal):
    """
    Computes sets of local states for which the locking in the initial state
    ensures that `goal` is impossible to reach.

    Elements of the returned lists can be notably used as argument for methods
    :py:meth:`.Model.having` and :py:meth:`.lock`.

    :param goal: goal specification (e.g., ``"a=1"``)
    :type goal: str or list(str) or .Goal
    :keyword kwgoal: keywords for goal specification (instead of `goal` argument)
    :keyword int maxsize: maximal cardinality of returned sets.
    :keyword set(str) exclude:
        set/list of automata to exclude from the solutions
    :keyword bool exclude_goal_automata:
        exclude automata involved in the goal specification
    :param int timeout: command timeout in seconds
    :rtype: list(dict[str,int])
    """
    goal = Goal.from_arg(goal, **kwgoal)

    info("This computation is an *under-approximation*: returned mutations \
are all valid, but they may be non-minimal, and some solutions may be missed.")
    info("Limiting solutions to mutations of at most %s automata. Use `maxsize` argument to change." % maxsize)

    args = ["--oneshot-mutations-for-cut", str(maxsize)]

    if isinstance(exclude, str):
        exclude = set([exclude])
    else:
        exclude = set(exclude)
    if exclude_goal_automata:
        exclude.update(goal.automata)
    if exclude:
        args += ["--ignore-automata", ",".join(['"%s"' % a for a in exclude])]

    cp = _run_tool("pint-reach", goal, *args, input_model=self, timeout=timeout)
    output = cp.stdout.decode()
    return json.loads(output)

@modeltool
def bifurcations(self, goal=None, method="ua", timeout=None, **kwgoal):
    """
    Identify local transitions after which, in some state, `goal` is no longer reachable.

    :param goal: goal specification (e.g., ``"a=1"``)
    :type goal: str or list(str) or .Goal
    :keyword kwgoal: keywords for goal specification (instead of `goal` argument)
    :keyword str method:

        * ``"exact"``, ``"nusmv"``, or ``"its"`` for complete identification of bifurcation
            transitions (PSPACE); ``"exact"`` is an alias for ``"nusmv"``.
        * ``"ua+mole"`` for under-approximation relying on exact the reachable
          states set prior computation (NP+PSPACE)
        * ``"ua"`` for under-approximation of bifurcation transitions (NP)
    :param int timeout: command timeout in seconds
    :rtype: :py:class:`.LocalTransition` list
    """
    assert method in ["exact", "ua", "mole+ua", "nusmv", "its"]

    goal = Goal.from_arg(goal, **kwgoal)

    if method == "exact":
        method = "nusmv"
    if method == "nusmv":
        smv = self.to_nusmv(skip_init=False)
        for tr in self.local_transitions:
            smv.add_ctl(tr.ctl_bifurcation(goal))
        r = smv.verify(as_dict=False, timeout=timeout)
        return [tr for (tr, r) in zip(self.local_transitions, r) if r]
    elif method == "its":
        itsm = self.to_its()
        ctls = [tr.ctl_bifurcation(goal) for tr in self.local_transitions]
        r = itsm.verify_ctls(ctls, timeout=timeout)
        return [tr for (tr, r) in zip(self.local_transitions, r) if r]
    else:
        info("This computation is an *under-approximation*: \
returned transitions are all bifurcation transitions, but some may have been missed. \
Use `method=\"exact\"` for complete identification.")
        cmd = "pint-reach"
        args = ["--bifurcations",
            "--bifurcations-method", method]

    cp = _run_tool(cmd, goal, *args, input_model=self, timeout=timeout)
    output = cp.stdout.decode()
    return [local_transition_from_json(d) for d in json.loads(output)]


__reachability_tools = ["its", "nusmv", "mole"]

@modeltool
def reachability(self, goal=None, fallback="its", tool="sa",
        sa_args=[],
        reduce_for_goal=True, timeout=None, **kwgoal):
    """
    Check if `goal` is reachable from the initial state.
    At first, Pint tries static analysis for the verification. If
    non-conclusive, it can fallback to exact model-checking.

    :param goal: goal specification (e.g., ``"a=1"``)
    :type goal: str or list(str) or .Goal
    :keyword kwgoal: keywords for goal specification (instead of `goal` argument)
    :keyword fallback: fallback to exact model-checking if static analysis
        is not conclusive. Supported model-checkers are: ``"its"``, ``"nusmv"``,
        and ``"mole"`` or a python function. Can be a list of fallbacks.
    :type fallback: str or list(str)
    :keyword str tool: tool for the model-checking:
        * ``"sa"``: static analysis with potential fallback method if not
        conclusive
        * ``"its"``, ``"nusmv"``, ``"mole"``: directly use the specified
        model-checker.
    :keyword list(str) sa_args: additional arguments for static analysis
    :keyword bool reduce_for_goal: before invoking a model-checker, perform the
        goal-oriented reduction of the automata network
    :param int timeout: command timeout in seconds
    :returns:

        * ``True`` if `goal` is reachable from :py:attr:`.initial_state`
        * ``False`` if `goal` is not reachable from :py:attr:`.initial_state`
        * :py:class:`.Inconc` if the static analysis is not conclusive and
          `fallback` is ``None``.
    """
    goal = Goal.from_arg(goal, **kwgoal)

    fallbacks = fallback if isinstance(fallback, list) else []
    if fallback is not None and not isinstance(fallback, list):
        fallbacks.append(fallback)

    chain = [tool] + fallbacks

    first_run = True
    for tool in chain:
        if tool == "sa":
            cp = _run_tool("pint-reach", goal, *sa_args, input_model=self, timeout=timeout)
            output = cp.stdout.decode()
            output = ternary(json.loads(output))
        else:
            if first_run:
                if reduce_for_goal:
                    self = self.reduce_for_goal(goal, squeeze_preserve=goal.automata)
                first_run = False
            if tool == "nusmv":
                smv = self.to_nusmv(skip_init=False)
                smv.add_ctl(EF(goal.to_ctl()))
                output = smv.alltrue(timeout=timeout)
            elif tool == "its":
                itsm = self.to_its()
                spec = goal.to_ctl()
                if goal.is_state_formula():
                    output = itsm.reachability(spec, timeout=timeout)
                else:
                    output = itsm.verify_ctl(spec, timeout=timeout)
            elif tool == "its-reach":
                m, mgoal = self.inject_goal(goal)
                mgoal = Goal.from_arg(mgoal)
                itsm = m.to_its()
                spec = mgoal.to_ctl()
                output = itsm.reachability(spec, timeout=timeout)
            elif tool == "mole":
                cp = _run_tool("pint-%s" % tool, goal, input_model=self,
                                timeout=timeout)
                output = cp.stdout.decode()
                output = ternary(json.loads(output))
            else:
                output = tool(self, goal)
        if output != Inconc:
            return output
        dbg(f"{tool} is inconclusive...")
    return Inconc

#TODO requirements


#
# pint-lcg
#

@modeltool
def local_causality_graph(self, kind="full", goal=None, **kwgoal):
    """
    Computes the Local Causality Graph (LCG) of type `kind`.

    :keyword str kind:

        * ``"full"``: LCG for all possible objectives from all possible initial
          states.
        * ``"verbose"``: LCG for simple over-approximation of `goal` rechability
        * ``"trimmed"``: LCG for simple over-approximation of `goal`
          reachability where impossible objectives have been removed.
        * ``"saturated"``: LCG for under-approximation of `goal` reachability
        * ``"worth"``: LCG used for `goal`-oriented reduction (see
          :py:meth:`.reduce`)

    :param goal: goal specification (e.g., ``"a=1"``) when `kind` is not ``"full"``.
    :type goal: str or list(str) or .Goal
    :keyword kwgoal: keywords for goal specification (instead of `goal` argument)
    :rtype: NetworkX multigraph (`nx.MultiDiGraph <http://networkx.readthedocs.io/en/stable/reference/classes.multidigraph.html>`_)

    .. seealso: methods :py:meth:`.full_lcg`, :py:meth:`.simple_lcg`, :py:meth:`.worth_lcg`, :py:meth:`.saturated_lcg`
    """
    assert kind in ["verbose","trimmed","saturated","worth","full"]
    goal = Goal.from_arg(goal, **kwgoal) if kind != "full" else None
    args = ["-t", kind, "-o", "-"]
    if goal:
        args.append(goal)
    cp = _run_tool("pint-lcg", *args, input_model=self)
    dot = StringIO(cp.stdout.decode())
    return nx.DiGraph(nx.nx_pydot.read_dot(dot))

@modeltool
def full_lcg(self):
    """
    Shortcut for :py:meth:`.local_causality_graph` with `kind="full"`
    """
    return local_causality_graph(self, "full")
@modeltool
def simple_lcg(self, goal=None, **kwgoal):
    """
    Shortcut for :py:meth:`.local_causality_graph` with `kind="trimmed"`
    """
    return local_causality_graph(self, "trimmed", goal=goal, **kwgoal)
@modeltool
def worth_lcg(self, goal, **kwgoal):
    """
    Shortcut for :py:meth:`.local_causality_graph` with `kind="worth"`
    """
    return local_causality_graph(self, "worth", goal=goal, **kwgoal)
@modeltool
def saturated_lcg(self, goal, **kwgoal):
    """
    Shortcut for :py:meth:`.local_causality_graph` with `kind="saturated"`
    """
    return local_causality_graph(self, "saturated", goal=goal, **kwgoal)


#
# pint-sg
#

@modeltool
def count_reachable_states(self, tool="pint", timeout=None):
    """
    Counts the exact number of states reachable from :py:attr:`.initial_state`.
    Uses an explicit state space approach.

    :keyword str tool:

        * ``"pint"`` explicit reachable state space computation (default);
        * ``"its"`` symbolic reable stae space computation
    :param int timeout: command timeout in seconds
    :rtype: int
    """
    assert tool in ["pint", "its"], "Wrong `tool` argument. See help."
    if tool == "pint":
        argv = ["pint-sg", "--count-reachable"]
        cp = _run_tool(*argv, input_model=self, timeout=timeout)
        output = cp.stdout.decode()
        return json.loads(output)
    else:
        return self.to_its().count_reachable(timeout=timeout)

@modeltool
def summary(model):
    """
    Returns a dictionnary with various information on the AN model:

    * ``"nb_automata"``: number of automata (equivalent to `len(m.automata)`)
    * ``"nb_local_states"``: total number of local states
    * ``"max_local_states"``: largest number of local states within one automaton.
    * ``"nb_transitions"``: number of local transitions
    * ``"nb_states"``: total number of states

    :rtype: dict
    """
    cp = _run_tool("pint-sg", "--description", input_model=model)
    output = cp.stdout.decode()
    return json.loads(output)

@modeltool
def reachable_states(self, timeout=None):
    """
    Returns the list of states reachable from :py:attr:`.initial_state`.

    :param int timeout: command timeout in seconds
    :rtype: state list
    """
    cp = _run_tool("pint-sg", "--reachable-states", input_model=self,
            timeout=timeout)
    output = cp.stdout.decode()
    return json.loads(output)

@modeltool
def reachable_stategraph(self, timeout=None):
    """
    Returns the reachable state graph from :py:attr:`.initial_state`.

    :param int timeout: command timeout in seconds
    :rtype: NetworkX digraph (`nx.DiGraph <http://networkx.readthedocs.io/en/stable/reference/classes.digraph.html>`_)
    """
    dotfile = new_output_file(ext="dot")
    _run_tool("pint-sg", "--state-graph", dotfile,
                input_model=self, stdout=None, timeout=timeout)
    g = nx.DiGraph(nx.nx_pydot.read_dot(dotfile))
    os.unlink(dotfile)
    return g

@modeltool
def reachable_attractors(self, timeout=None):
    """
    Returns the complete list of attractors reachable from
    :py:attr:`.initial_state`.

    Uses an explicit state space exploration methods.

    :param int timeout: command timeout in seconds

    Each attractor is described by a `dict` object with the following keys:

    * ``"type"``: either ``"fixpoint"`` or ``"cyclic"``.
    * ``"size"``: number of states in the attractor (1 if fixpoint, >1 if
      cyclic).
    * ``"sample"``: state (represented as `dict`) belonging to the attractor,
      i.e., either the fixpoint, or one of the state in the cycle attractor.
    """
    cp = _run_tool("pint-sg", "--reachable-attractors", input_model=self,
            timeout=timeout)
    output = cp.stdout.decode()
    return json.loads(output)

#
# pint-stable
#

@modeltool
def fixpoints(self, timeout=None):
    """
    Returns the complete list of fixed points of the model.

    :param int timeout: command timeout in seconds
    :rtype: dict list
    """
    cp = _run_tool("pint-stable", "--fixpoints", input_model=self,
            timeout=timeout)
    output = cp.stdout.decode()
    return json.loads(output)


#
# misc
#
@modeltool
def dependency_graph(self):
    """
    Returns the dependency graph between automata:
    there is an edge from `a` to `b` if some local transitions of `b` depends on `a`.

    Complexity: linear with the number of local transitions.

    :rtype: NetworkX digraph (`nx.DiGraph <http://networkx.readthedocs.io/en/stable/reference/classes.digraph.html>`_)
    """
    g = nx.DiGraph()
    g.add_nodes_from(self.automata)
    for tr in self.local_transitions:
        for a in tr.modified_automata:
            for b in tr.conds.keys():
                g.add_edge(b, a)
    return g


@modeltool
def automaton_graph(self, a):
    """
    Returns the directed graph of local transitions between the local states of
    automaton `a`. Edges are labeled with the index of the transitions in
    :py:attr:`.local_transitions`.

    :rtype: NetworkX digraph (`nx.DiGraph <http://networkx.readthedocs.io/en/stable/reference/classes.digraph.html>`_)
    """
    g = nx.DiGraph()
    g.add_nodes_from(self.local_states[a])
    for (idx, tr) in enumerate(self.local_transitions):
        if a in tr.modified_automata:
            if isinstance(tr, LocalTransition):
                i, j = tr.i, tr.j
            else:
               ((i, j),) = [(i,j) for (b,i,j) in tr.local_transitions if b == a]
            g.add_edge(i, j, label=str(idx))
    return g

@modeltool
def ctl_of_cutset(self, cutset, goal=None, **kwgoal):
    """
    Return the CTL specification for `cutset` being a cut-set for `goal`
    reachability.

    :param goal: goal specification (e.g., ``"a=1"``)
    :type goal: str or list(str) or .Goal
    :keyword kwgoal: keywords for goal specification (instead of `goal` argument)
    :param cutset: cut-set
    :type cutset: list(dict[str,int or int list])
    """
    goal = Goal.from_arg(goal, **kwgoal)
    def ls_ctl(ai):
        (a,i) = ai
        if isinstance(i, int):
            return S({a:i})
        else:
            return reduce(lambda x, y: x|y, [ls_ctl((a,j)) for j in i])
    cutset_ctl = reduce(lambda x, y: x|y, map(ls_ctl, cutset.items()))
    return ~EU(~cutset_ctl, goal.to_ctl())

__all__ = [t[0] for t in __MODEL_TOOLS] + [
    "EXPORT_SUPPORTED_FORMATS",
    "EXPORT_SUPPORTED_EXTENSIONS",
    "PintProcessError",
    "EquipTools",
    ]






def pint_protect(data):
    if isinstance(data, str):
        return '"%s"' % data
    else:
        return str(data)

def pint_of_ls(ai):
    return "{}={}".format(*map(pint_protect, ai))

def pint_of_localtransition(changes, conds):
    changes = ["{} {} -> {}".format(*map(pint_protect, c)) for c in changes]
    if len(changes) > 1:
        changes = "{ %s }" % " ; ".join(changes)
    else:
        changes = changes[0]
    conds = dict(conds)
    conds = " when %s" % " and ".join(map(pint_of_ls, conds.items())) if conds else ""
    return changes + conds





import os
import unittest

import pypint

__data_dir__ = os.path.join(os.path.dirname(__file__), "data")

class LoadGINsimTest(unittest.TestCase):
    def test_zginml(self):
        assert pypint.load(os.path.join(__data_dir__, "phageLambda4.zginml"))
    def test_sbml(self):
        assert pypint.load(os.path.join(__data_dir__, "phageLambda4.sbml"))





import base64
import copy
import json
import operator
import os
import random
import re
import subprocess
import tempfile
from urllib.parse import urlparse
from urllib.request import urlretrieve
from xml.dom.minidom import parse as xml_parse_dom
from zipfile import ZipFile

from colomoto_jupyter import import_colomoto_tool
from colomoto_jupyter.io import download, ensure_localfile

from .cfg import *
from .tools import *
from .types import *
from .ui import *
from .utils import *
from .converters import CONVERTERS

if IN_IPYTHON:
    from colomoto_jupyter.upload import jupyter_upload

class InitialState(dict):
    """
    A state is represented with a ``dict``-like python object, which associates
    each automaton to a local state, or to a set of initial local states.

    The automata, their value domain, as well as their initial *default* value,
    are tied to a :py:class:`.Model` object.

    A new `InitialState` should be instanciated only from an existing
    `InitialState` instance, typically :py:attr:`.Model.initial_state`,
    either using the copy constructor, or using :py:meth:`.InitialState.copy`,
    or using :py:meth:`.InitialState.having` method (recommended).
    """
    def __init__(self, state):
        if isinstance(state, InitialState):
            # copy constructor
            s = state
            for field in ["info", "defaults", "domain"]:
                setattr(self, field, getattr(s, field))
            self.__override = s.__override.copy()
            super(InitialState, self).__init__(self.defaults)
            return
        self.info = state
        self.defaults = self.info["initial_state"]
        super(InitialState, self).__init__(self.defaults)
        self.domain = {}
        for a in self.info["automata"]:
            self.domain[a] = set(self.info["local_states"][a])
            if type(self.defaults[a]) is list:
                self.defaults[a] = tuple(self.defaults[a])
        self.__override = {}

    def is_custom(self):
        """
        :returns: `True` iff the state is different from the default initial state of the AN.
        """
        return len(self.__override) > 0

    def __delitem__(self, a):
        """
        Restore the default initial value of automaton `a`.

        >>> del m.initial_state["a"]
        """
        del self.__override[a]
        super(InitialState, self).__setitem__(a, self.defaults[a])

    def __getitem__(self, key):
        return self.__override.get(key, self.defaults[key])

    def __setitem__(self, a, i):
        """
        Sets the initial state of automaton `a` to `i`.
        `i` can be either a single local state, or a set of local states.
        """
        key = a
        value = i
        if type(value) not in [int, str, list, set, tuple]:
            raise TypeError("Invalid type value")
        if key not in self.defaults:
            raise TypeError("Unknown automaton '%s'" % key)
        if type(value) in [int, str] and value in self.domain[key] \
                or self.domain[key].issuperset(set(value)):
            if type(value) not in [int, str]:
                value = tuple(value)
            if value == self.defaults[key]:
                if key in self.__override:
                    del self.__override[key]
            else:
                self.__override[key] = value
                super(InitialState, self).__setitem__(key, value)
        else:
            raise TypeError("Invalid value for '%s' (allowed: %s)" \
                                % (key, ", ".join(map(str, self.domain[key]))))

    def copy(self):
        """
        :returns: a copy of the object.
        """
        return InitialState(self)

    def having(self, *args, **kwargs):
        """
        :returns: a copy of the object modified with :py:meth:`.update` method.
        """
        s = self.copy()
        s.update(*args, **kwargs)
        return s

    def update(self, *args, **F):
        """
        Sets the initial local states of specified automata. Arguments can be
        either `dict`-like objects, lists of pairs or keywords.

        Examples:

        >>> s.update(a=1, b=0)  # updates automaton 'a' and 'b'
        >>> s.update({"a": 1, "b": 0})  # equivalent to above
        """
        for E in args:
            if hasattr(E, "keys"):
                for k in E:
                    self[k] = E[k]
            else:
                for k,v in E:
                    self[k] = v
        for k,v in F.items():
            self[k] = v

    def reset(self):
        """
        Restore to default initial state.
        """
        self.__override.clear()
        super(InitialState, self).update(self.defaults)

    def changes(self):
        """
        :returns: A `dict` object of the changes to the default initial state.
        """
        return self.__override.copy()

    def nonzeros(self):
        """
        :returns: A `dict` object associating automata to their initial local
            states when it is different from ``0``.
        """
        return dict([(a,i) for (a,i) in self.items() if isinstance(i, str) or i > 0])

    def to_pint(self, full=False):
        """
        :returns: Pint text representation of the state
        :keyword full: if `True`, also output initial states being ``0``.
        """
        items = self.items() if full else self.__override.items()
        return pint_of_localstates(items)


def InfoFields(*fields):
    def plug(cls):
        for field in fields:
            setattr(cls, field, property(lambda self, field=field: self.info[field]))
        return cls
    return plug

@EquipTools
@InfoFields("automata", "local_states", "features", "local_transitions")
class Model(object):
    """
    Abstract class for manipulating and analyzing an AN (Automata Network) model.

    A `Model` is typically instanciated using the :py:func:`.load` function, or
    using :py:class:`.InMemoryModel` for small in-memory AN specifications.

    .. py:attribute:: automata

        The list of automata names

    .. py:attribute:: local_states

        Dictionnary associating to each automaton its list of local states
        (`str` or `int` list).

    .. py:attribute:: local_transitions

        The list of local transitions (:py:class:`.LocalTransition` or
        :py:class:`.SynchronizedLocalTransitions`) in the AN specification.

    .. py:attribute:: initial_state

        The current initial state of the model (:py:class:`.InitialState`
        instance).

    .. py:attribute:: named_states

        A dictionnary of named :py:class:`.InitialState` instances.

    .. py:attribute:: features

        A list of features the AN model is using, if any, among:

        * `"synced_transitions"`: the AN uses synchronized local transitions.
    """
    def __init__(self):
        self.named_states = {}

    def load(self):
        args = ["pint-export", "-l", "nbjson"]
        kwargs = {}
        self.__initial_state = None
        self.populate_popen_args(args, kwargs)
        kwargs["stderr"]= subprocess.PIPE
        try:
            self.info = json.loads(subprocess.check_output(args, **kwargs).decode())
        except subprocess.CalledProcessError as e:
            # backward compatible 'raise e from None'
            e = PintProcessError(e.returncode, e.cmd, e.output, e.stderr)
            e.__cause__ = None
            raise e

        key = "local_transitions"
        if key in self.info:
            self.info[key] = [local_transition_from_json(d) \
                                    for d in self.info[key]]
        self.__initial_state = InitialState(self.info)

    def set_initial_state(self, state):
        assert state.info == self.info
        self.__initial_state = state
    initial_state = property(lambda self: self.__initial_state, set_initial_state)

    def populate_popen_args(self, args, kwargs):
        if self.initial_state is not None and self.initial_state.is_custom():
            args += ["--initial-context", self.initial_state.to_pint()]

    def register_state(self, name, state):
        """
        Register a named state in :py:attr:`named_states` attribute.

        :param str name: name of the state
        :param state: state to register
        :type state: dict or .InitialState
        """
        self.named_states[name] = InitialState(self.info)
        self.named_states[name].update(state)

    def having(self, initial_state=None, **kwargs):
        """
        Returns a copy of the model with supplied modifications

        :param initial_state: new initial state
        :type initial_state: .InitialState or str
        :keyword kwargs: if non-empty, defines a new initial state from
            :py:attr:`.Model.initial_state` by calling
            :py:meth:`.InitialState.having`.
        :rtype: :py:class:`.Model`

        Exemples:

        >>> m.having(m.named_states["ProT1"]).reachability("Tf1=1")
        >>> m.having(HR=1).reachability("Tf1=1")
        """
        m = copy.copy(self)
        if isinstance(initial_state, str):
            m.initial_state = m.named_states[initial_state]
            return m
        m.initial_state = self.initial_state.copy()
        args = []
        if initial_state:
            args.append(initial_state)
        m.initial_state.update(*args, **kwargs)
        return m


class FileModel(Model):
    """
    Concrete class of :py:class:`.Model` for AN model stored in a local file.
    """
    def __init__(self, filename):
        super(FileModel, self).__init__()
        self.filename = filename
        self.load()

    def populate_popen_args(self, args, kwargs):
        args += ["-i", self.filename]
        super(FileModel, self).populate_popen_args(args, kwargs)

    def source(self):
        """
        Returns the text source in Pint native format
        """
        with open(self.filename) as f:
            data = f.read()
        return data

class InMemoryModel(Model):
    """
    Concrete class of :py:class:`.Model` for an AN model stored in memory.
    """
    def __init__(self, data):
        super(InMemoryModel, self).__init__()
        self.data = data if type(data) is bytes else data.encode("utf-8")
        self.load()

    def populate_popen_args(self, args, kwargs):
        kwargs["input"] = self.data
        super(InMemoryModel, self).populate_popen_args(args, kwargs)

    def source(self):
        """
        Returns the text source in Pint native format
        """
        return self.data



def import_with_ginsim(fmt, inputfile, anfile, simplify=True):

    # some file formats, such as zginml, can define named states
    states = {}

    if fmt == "zginml":
        def parse_localstate(value):
            a,i = value.split(";")
            return (a,int(i))

        def parse_state(value):
            return dict(map(parse_localstate, value.strip().split()))

        initialState_e = "GINsim-data/initialState"
        with ZipFile(inputfile) as z:
            if initialState_e in z.namelist():
                with z.open("GINsim-data/initialState") as f:
                    dom = xml_parse_dom(f)
                    for state in dom.getElementsByTagName("initialState"):
                        name = state.getAttribute("name")
                        states[name] = parse_state(state.getAttribute("value"))

    biolqm = import_colomoto_tool("biolqm")
    lqm = biolqm.load(inputfile)
    model = biolqm.to_pint(lqm, simplify=simplify)

    if states:
        for name, state in states.items():
            model.register_state(name, state)
        info("%d state(s) have been registered: %s" \
            % (len(states), ", ".join(states.keys())))

    return model


ext2format = {
    "an": "an",
    "bc": "biocham",
    "bcx": "cadbiom",
    "bn": "boolfunctions",
    "booleannet": "booleannet",
    "boolfunctions": "boolfunctions",
    "boolsim": "boolsim",
    "ginml": "ginml",
    "sbgn": "sbgnpd",
    "sbgnml": "sbgnpd",
    "sbml": "sbml",
    "zginml": "zginml",
}

LOAD_SUPPORTED_FORMATS = list(sorted(set(ext2format.values())))
"""
Formats supported by :py:func:`.load` method
"""


def load(filename=None, format=None, simplify=True, **opts):
    """
    Load a Pint model from given filename.
    The format is guessed from the filename extension, but can be enforced with
    the `format` parameter.
    Except when loading directly from a ``.an`` file (native format for Pint), the
    model will be converted to .an and a simplification step is performed.
    This latter stage can be deactivated with `simplify=False`

    `filename` can be either a path to local file, or an URL. In the latter
    case, it will be downloaded locally first, and then processed as a local
    file.

    Within IPython/Jupyter web notebook, the `filename` argument can be omited:
    in that case, a file upload form will be displayed.

    Supported file extensions:

    * ``.an`` (automata network), native Pint file format;
    * ``.ginml``, imported using GINsim (require `ginsim` Python module)
    * ``.zginml``, like ginml, with in addition the support of named initial states;
    * ``.sbml`` (SBML-qual), imported using GINsim (require `ginsim` Python module);
    * ``.boolsim``, ``.booleannet``, ``.boolfunctions`` (or ``.bn``): Boolean network formats, imported
      using GINsim (require `ginsim` Python module)`.
    * ``.bc``: Biocham model - the importation implements its Boolean semantics
      (experimental feature)

    Supported URL schemes:

    * `CellCollective <htts://cellcollective.org>`_ models can be imported by
      supplying the URL of the model main page. The SBML file location is
      determined automatically;
    * any other URL is treated as a link to a file with supported extension. It
      will be downloaded and then loaded as a local file.

    Returns a :py:class:`.Model` instance.
    If the model results from an importation, IPython displays the link to the
    generated .an file.

    Examples:

    >>> m1 = pypint.load("mylocalfile.an")
    >>> m2 = pypint.load("http://ginsim.org/sites/default/files/Frontiers-Th-Full-model-annotated.zginml")
    >>> m3 = pypint.load("https://cellcollective.org/#4705/septation-initiation-network")

    .. seealso:: :py:data:`.LOAD_SUPPORTED_FORMATS` for supported formats.
    """

    if filename is None:
        if IN_IPYTHON:
            return jupyter_upload("pypint.load", "pypint.load")
        else:
            raise TypeError("missing filename argument")

    filename = ensure_localfile(filename)

    bname = os.path.basename(filename)
    ext = file_ext(filename)
    name = bname[:-len(ext)-1][-15:]
    if format is None:
        assert ext in ext2format, "Unknown extension '%s'" % ext
        format = ext2format[ext]

    def make_anfile():
        return new_output_file(suffix="%s.an"%name)

    if format == "an":
        dbg("Source file is in Automata Network (an) format")
        return FileModel(filename, **opts)

    elif format in ["boolfunctions", "boolsim", "booleannet", "sbml",
                    "ginml", "zginml"]:
        dbg("Source file is in %s format, importing with GINsim" \
                    % format)
        anfile = make_anfile()
        return import_with_ginsim(format, filename, anfile,
                    simplify=simplify, **opts)

    elif format in CONVERTERS:
        dbg("Source file is in %s format" % format)
        anfile = make_anfile()
        with open(anfile, "w") as outfd:
            CONVERTERS[format](filename, outfd, **opts)
        return FileModel(anfile)
    else:
        raise ValueError("Format '%s' is not supported." % format)


__all__ = [
    "LOAD_SUPPORTED_FORMATS",
    "load",
    "Model", "FileModel", "InMemoryModel",
    "InitialState"]





import os
import sys
from argparse import ArgumentParser

from .model import load, FileModel

from .cfg import CFG, remove_output_files

def import_model():
    CFG["console"] = True
    parser = ArgumentParser(prog=sys.argv[0])
    parser.add_argument("input")
    parser.add_argument("--output", "-o", type=str, default=None)

    sbgnpd_opts = parser.add_argument_group("SBGN-PD import")
    sbgnpd_opts.add_argument("--fancy-names", action="store_true", default=False,
                help="Use EPN labels as automata names")

    args = parser.parse_args()
    opts = vars(args).copy()
    del opts["input"]
    del opts["output"]

    an = load(args.input, **opts)

    if args.output:
        an.save_as(args.output)
    else:
        print(an.source())

    remove_output_files()






from colomoto_jupyter.sessionfiles import new_output_file
from pypint.converters.lib.boolean_utils import BoolToAN
from pypint.converters.lib.export_utils import pint_protect

from colomoto.minibn import BooleanNetwork, MultiValuedNetwork

def import_minibn(f):
    if isinstance(f, MultiValuedNetwork):
        raise Exception("Direct conversion from MultiValuedNetwork is not supported yet (you should use biolqm instead)")

    assert isinstance(f, BooleanNetwork), \
        "{}: only objects of type {} are supported".format(import_minibn, BooleanNetwork)

    ba = f.ba

    def ls_of_lit(lit):
        if isinstance(lit, ba.NOT):
            return (lit.args[0].obj, 0)
        else:
            return (lit.obj, 1)

    anfile = new_output_file(ext="an")
    with open(anfile, "w") as outf:
        def out(data):
            print(data, file=outf)
        b2a = BoolToAN(ba, ls_of_lit, out)

        for a in sorted(f.keys()):
            out("{} [0, 1]".format(pint_protect(a)))

        for a, fa in sorted(f.items()):
            sa = f.vars(a)[0]
            expr_up = fa.subs({sa: ba.FALSE}).literalize().simplify()
            expr_down = (~fa).subs({sa: ba.TRUE}).literalize().simplify()
            b2a.make_transitions([(a, 0, 1)], expr_up)
            b2a.make_transitions([(a, 1, 0)], expr_down)

    import pypint
    return pypint.load(anfile)

